#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

"""The secret detector detects sensitive information.

It masks secrets that might be leaked from two potential avenues
    1. Out of Band Telemetry
    2. Logging
"""
from __future__ import annotations

import logging
import os
import re

MIN_TOKEN_LEN = os.getenv("MIN_TOKEN_LEN", 32)
MIN_PWD_LEN = os.getenv("MIN_PWD_LEN", 8)


class SecretDetector(logging.Formatter):
    AWS_KEY_PATTERN = re.compile(
        r"(aws_key_id|aws_secret_key|access_key_id|secret_access_key)\s*=\s*'([^']+)'",
        flags=re.IGNORECASE,
    )
    AWS_TOKEN_PATTERN = re.compile(
        r'(accessToken|tempToken|keySecret)"\s*:\s*"([a-z0-9/+]{32,}={0,2})"',
        flags=re.IGNORECASE,
    )
    SAS_TOKEN_PATTERN = re.compile(
        r"(sig|signature|AWSAccessKeyId|password|passcode)=(?P<secret>[a-z0-9%/+]{16,})",
        flags=re.IGNORECASE,
    )
    PRIVATE_KEY_PATTERN = re.compile(
        r"-----BEGIN PRIVATE KEY-----\\n([a-z0-9/+=\\n]{32,})\\n-----END PRIVATE KEY-----",
        flags=re.MULTILINE | re.IGNORECASE,
    )
    PRIVATE_KEY_DATA_PATTERN = re.compile(
        r'"privateKeyData": "([a-z0-9/+=\\n]{10,})"', flags=re.MULTILINE | re.IGNORECASE
    )
    CONNECTION_TOKEN_PATTERN = re.compile(
        r"(token|assertion content)" r"([\'\"\s:=]+)" r"([a-z0-9=/_\-\+]{8,})",
        flags=re.IGNORECASE,
    )

    PASSWORD_PATTERN = re.compile(
        r"(password"
        r"|pwd)"
        r"([\'\"\s:=]+)"
        r"([a-z0-9!\"#\$%&\\\'\(\)\*\+\,-\./:;<=>\?\@\[\]\^_`\{\|\}~]{1,})",
        flags=re.IGNORECASE,
    )

    @staticmethod
    def mask_connection_token(text: str) -> str:
        return SecretDetector.CONNECTION_TOKEN_PATTERN.sub(r"\1\2****", text)

    @staticmethod
    def mask_password(text: str) -> str:
        return SecretDetector.PASSWORD_PATTERN.sub(r"\1\2****", text)

    @staticmethod
    def mask_aws_keys(text: str) -> str:
        return SecretDetector.AWS_KEY_PATTERN.sub(r"\1='****'", text)

    @staticmethod
    def mask_sas_tokens(text: str) -> str:
        return SecretDetector.SAS_TOKEN_PATTERN.sub(r"\1=****", text)

    @staticmethod
    def mask_aws_tokens(text: str) -> str:
        return SecretDetector.AWS_TOKEN_PATTERN.sub(r'\1":"XXXX"', text)

    @staticmethod
    def mask_private_key(text: str) -> str:
        return SecretDetector.PRIVATE_KEY_PATTERN.sub(
            "-----BEGIN PRIVATE KEY-----\\\\nXXXX\\\\n-----END PRIVATE KEY-----", text
        )

    @staticmethod
    def mask_private_key_data(text: str) -> str:
        return SecretDetector.PRIVATE_KEY_DATA_PATTERN.sub(
            '"privateKeyData": "XXXX"', text
        )

    @staticmethod
    def mask_secrets(text: str) -> tuple[bool, str, str | None]:
        """Masks any secrets. This is the method that should be used by outside classes.

        Args:
            text: A string which may contain a secret.

        Returns:
            The masked string.
        """
        if text is None:
            return (False, None, None)

        masked = False
        err_str = None
        try:
            masked_text = SecretDetector.mask_connection_token(
                SecretDetector.mask_password(
                    SecretDetector.mask_private_key_data(
                        SecretDetector.mask_private_key(
                            SecretDetector.mask_aws_tokens(
                                SecretDetector.mask_sas_tokens(
                                    SecretDetector.mask_aws_keys(text)
                                )
                            )
                        )
                    )
                )
            )
            if masked_text != text:
                masked = True
        except Exception as ex:
            # We'll assume that the exception was raised during masking
            # to be safe consider that the log has sensitive information
            # and do not raise an exception.
            masked = True
            masked_text = str(ex)
            err_str = str(ex)

        return masked, masked_text, err_str

    def format(self, record: logging.LogRecord) -> str:
        """Wrapper around logging module's formatter.

        This will ensure that the formatted message is free from sensitive credentials.

        Args:
            record: The logging record.

        Returns:
            Formatted desensitized log string.
        """
        try:
            unsanitized_log = super().format(record)
            masked, sanitized_log, err_str = SecretDetector.mask_secrets(
                unsanitized_log
            )
            if masked and err_str is not None:
                sanitized_log = "{} - {} {} - {} - {} - {}".format(
                    record.asctime,
                    record.threadName,
                    "secret_detector.py",
                    "sanitize_log_str",
                    record.levelname,
                    err_str,
                )
        except Exception as ex:
            sanitized_log = "{} - {} {} - {} - {} - {}".format(
                record.asctime,
                record.threadName,
                "secret_detector.py",
                "sanitize_log_str",
                record.levelname,
                "EXCEPTION - " + str(ex),
            )
        return sanitized_log



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import os
import sys
from os import path
from typing import TYPE_CHECKING

from snowflake.connector.ocsp_asn1crypto import SnowflakeOCSPAsn1Crypto

if TYPE_CHECKING:
    from asn1crypto.x509 import Certificate


def main() -> None:
    """Internal Tool: Extract certificate files in PEM."""

    def help() -> None:
        print(
            "Extract certificate file. The target file can be a single file "
            "or a directory including multiple certificates. The certificate "
            "file format should be PEM."
        )
        print(
            """
Usage: {}  <input file/dir>
""".format(
                path.basename(sys.argv[0])
            )
        )
        sys.exit(2)

    if len(sys.argv) < 2:
        help()

    input_filename = sys.argv[1]
    if path.isdir(input_filename):
        files = [path.join(input_filename, f) for f in os.listdir(input_filename)]
    else:
        files = [input_filename]

    for f in files:
        open(f)
        extract_certificate_file(f)


def extract_certificate_file(input_filename) -> None:
    ocsp = SnowflakeOCSPAsn1Crypto()
    cert_map: dict[bytes, Certificate] = {}
    ocsp.read_cert_bundle(input_filename, cert_map)

    for cert in cert_map.values():
        print(f"serial #: {cert.serial_number}, name: {cert.subject.native}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

from enum import Enum
from unittest import mock

from snowflake.connector.network import SnowflakeRestful

try:
    from snowflake.connector.ssl_wrap_socket import DEFAULT_OCSP_MODE
except ImportError:

    class OCSPMode(Enum):
        FAIL_OPEN = "FAIL_OPEN"

    DEFAULT_OCSP_MODE = OCSPMode.FAIL_OPEN

hostname_1 = "sfctest0.snowflakecomputing.com"
url_1 = f"https://{hostname_1}:443/session/v1/login-request"

hostname_2 = "sfc-ds2-customer-stage.s3.amazonaws.com"
url_2 = f"https://{hostname_2}/rgm1-s-sfctest0/stages/"
url_3 = f"https://{hostname_2}/rgm1-s-sfctst0/stages/another-url"


mock_conn = mock.Mock()
mock_conn.disable_request_pooling = False
mock_conn._ocsp_mode = lambda: DEFAULT_OCSP_MODE


def close_sessions(rest: SnowflakeRestful, num_session_pools: int) -> None:
    """Helper function to call SnowflakeRestful.close(). Asserts close was called on all SessionPools."""
    with mock.patch("snowflake.connector.network.SessionPool.close") as close_mock:
        rest.close()
        assert close_mock.call_count == num_session_pools


def create_session(
    rest: SnowflakeRestful, num_sessions: int = 1, url: str | None = None
) -> None:
    """
    Creates 'num_sessions' sessions to 'url'. This is recursive so that idle sessions
    are not reused.
    """
    if num_sessions == 0:
        return
    with rest._use_requests_session(url):
        create_session(rest, num_sessions - 1, url)


@mock.patch("snowflake.connector.network.SnowflakeRestful.make_requests_session")
def test_no_url_multiple_sessions(make_session_mock):
    rest = SnowflakeRestful(connection=mock_conn)

    create_session(rest, 2)

    assert make_session_mock.call_count == 2

    assert list(rest._sessions_map.keys()) == [None]

    session_pool = rest._sessions_map[None]
    assert len(session_pool._idle_sessions) == 2
    assert len(session_pool._active_sessions) == 0

    close_sessions(rest, 1)


@mock.patch("snowflake.connector.network.SnowflakeRestful.make_requests_session")
def test_multiple_urls_multiple_sessions(make_session_mock):
    rest = SnowflakeRestful(connection=mock_conn)

    for url in [url_1, url_2, None]:
        create_session(rest, num_sessions=2, url=url)

    assert make_session_mock.call_count == 6

    hostnames = list(rest._sessions_map.keys())
    for hostname in [hostname_1, hostname_2, None]:
        assert hostname in hostnames

    for pool in rest._sessions_map.values():
        assert len(pool._idle_sessions) == 2
        assert len(pool._active_sessions) == 0

    close_sessions(rest, 3)


@mock.patch("snowflake.connector.network.SnowflakeRestful.make_requests_session")
def test_multiple_urls_reuse_sessions(make_session_mock):
    rest = SnowflakeRestful(connection=mock_conn)
    for url in [url_1, url_2, url_3, None]:
        # create 10 sessions, one after another
        for _ in range(10):
            create_session(rest, url=url)

    # only one session is created and reused thereafter
    assert make_session_mock.call_count == 3

    hostnames = list(rest._sessions_map.keys())
    assert len(hostnames) == 3
    for hostname in [hostname_1, hostname_2, None]:
        assert hostname in hostnames

    for pool in rest._sessions_map.values():
        assert len(pool._idle_sessions) == 1
        assert len(pool._active_sessions) == 0

    close_sessions(rest, 3)



#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from unittest import mock

from snowflake.connector import connect


def test_mfa_token_cache():
    with mock.patch(
        "snowflake.connector.network.SnowflakeRestful.fetch",
    ):
        with mock.patch(
            "snowflake.connector.auth._auth.Auth._write_temporary_credential",
        ) as save_mock:
            with connect(
                account="account",
                user="user",
                password="password",
                authenticator="username_password_mfa",
                client_store_temporary_credential=True,
                client_request_mfa_token=True,
            ):
                assert save_mock.called
    with mock.patch(
        "snowflake.connector.network.SnowflakeRestful.fetch",
        return_value={
            "data": {
                "token": "abcd",
                "masterToken": "defg",
            },
            "success": True,
        },
    ):
        with mock.patch(
            "snowflake.connector.cursor.SnowflakeCursor._init_result_and_meta",
        ):
            with mock.patch(
                "snowflake.connector.auth._auth.Auth._read_temporary_credential",
                return_value=None,
            ) as load_mock:
                with connect(
                    account="account",
                    user="user",
                    password="password",
                    authenticator="username_password_mfa",
                    client_store_temporary_credential=True,
                    client_request_mfa_token=True,
                ):
                    assert load_mock.called



#!/usr/bin/env python
#
# Copyright (c) 2012-2023 Snowflake Computing Inc. All rights reserved.
#

from __future__ import annotations

import time
from concurrent.futures import ThreadPoolExecutor

from snowflake.connector.errors import RevocationCheckError
from snowflake.connector.telemetry_oob import TelemetryService

TEST_RACE_CONDITION_THREAD_COUNT = 2
TEST_RACE_CONDITION_DELAY_SECONDS = 1
telemetry_data = {}
exception = RevocationCheckError("Test OCSP Revocation error")
event_type = "Test OCSP Exception"
stack_trace = [
    "Traceback (most recent call last):\n",
    '  File "<doctest...>", line 10, in <module>\n    lumberjack()\n',
    '  File "<doctest...>", line 4, in lumberjack\n    bright_side_of_death()\n',
    '  File "<doctest...>", line 7, in bright_side_of_death\n    return tuple()[0]\n',
    "IndexError: tuple index out of range\n",
]

event_name = "HttpRetryTimeout"
url = "http://localhost:8080/queries/v1/query-request?request_guid=a54a3d70-abf2-4576-bb6f-ddf23999491a"
method = "POST"


def test_telemetry_oob_disabled():
    telemetry = TelemetryService.get_instance()
    assert not telemetry.enabled
    telemetry.enable()
    assert not telemetry.enabled
    telemetry.disable()
    assert not telemetry.enabled
    telemetry.enable()
    telemetry.log_ocsp_exception(
        event_type, telemetry_data, exception=exception, stack_trace=stack_trace
    )
    assert telemetry.queue.empty()
    telemetry.log_general_exception(event_name, {})
    assert telemetry.queue.empty()
    telemetry.log_http_request_error(event_name, url, method, "error", "error")
    assert telemetry.queue.empty()


class MockTelemetryService(TelemetryService):
    """Mocks a delay in the __init__ of TelemetryService to simulate a race condition"""

    def __init__(self, *args, **kwargs):
        # this delay all but guarantees enough time to catch multiple threads entering __init__
        time.sleep(TEST_RACE_CONDITION_DELAY_SECONDS)
        super().__init__(*args, **kwargs)


def test_get_instance_multithreaded():
    """Tests thread safety of multithreaded calls to TelemetryService.get_instance()"""
    TelemetryService._TelemetryService__instance = None
    with ThreadPoolExecutor() as executor:
        futures = [
            executor.submit(MockTelemetryService.get_instance)
            for _ in range(TEST_RACE_CONDITION_THREAD_COUNT)
        ]
        for future in futures:
            # will error if singleton constraint violated
            future.result()
